@@ -42,8 +42,10 @@ from gi.repository import Gtk, Gdk, GLib, Gio, GObject, Gst, GstAudio, GstPbutil
 from gettext import gettext as _, ngettext
 from random import randint
 from collections import deque
-from gnomemusic.albumArtCache import AlbumArtCache
+from gnomemusic.albumartcache import AlbumArtCache, DefaultIcon, ArtSize
+from gnomemusic.grilo import grilo
 from gnomemusic.playlists import Playlists
+import gnomemusic.utils as utils
 playlists = Playlists.get_default()
 
 from hashlib import md5
@@ -55,8 +57,6 @@ from gnomemusic import log
 import logging
 logger = logging.getLogger(__name__)
 
-ART_SIZE = 34
-
 
 class RepeatType:
     NONE = 0
@@ -80,6 +80,7 @@ class DiscoveryStatus:
 class Player(GObject.GObject):
     nextTrack = None
     timeout = None
+    _seconds_timeout = None
     shuffleHistory = deque(maxlen=10)
 
     __gsignals__ = {
@@ -92,7 +93,7 @@ class Player(GObject.GObject):
         'volume-changed': (GObject.SignalFlags.RUN_FIRST, None, ()),
         'prev-next-invalidated': (GObject.SignalFlags.RUN_FIRST, None, ()),
         'seeked': (GObject.SignalFlags.RUN_FIRST, None, (int,)),
-        'thumbnail-updated': (GObject.SignalFlags.RUN_FIRST, None, (str,)),
+        'thumbnail-updated': (GObject.SignalFlags.RUN_FIRST, None, ()),
     }
 
     def __repr__(self):
@@ -108,10 +109,11 @@ class Player(GObject.GObject):
         self.playlistField = None
         self.currentTrack = None
         self.currentTrackUri = None
-        self._lastState = Gst.State.PAUSED
-        self.cache = AlbumArtCache.get_default()
-        self._noArtworkIcon = self.cache.get_default_icon(ART_SIZE, ART_SIZE)
-        self._loadingIcon = self.cache.get_default_icon(ART_SIZE, ART_SIZE, True)
+        scale = parent_window.get_scale_factor()
+        self.cache = AlbumArtCache(scale)
+        self._loading_icon_surface = DefaultIcon(scale).get(
+            DefaultIcon.Type.loading,
+            ArtSize.xsmall)
         self._missingPluginMessages = []
 
         Gst.init(None)
@@ -240,6 +242,7 @@ class Player(GObject.GObject):
         self.repeat = settings.get_enum('repeat')
         self._sync_prev_next()
         self._sync_repeat_image()
+        self._validate_next_track()
 
     @log
     def _on_bus_state_changed(self, bus, message):
@@ -370,7 +373,7 @@ class Player(GObject.GObject):
         elif (self.repeat == RepeatType.NONE):
             self.stop()
             self.playBtn.set_image(self._playImage)
-            self.progressScale.set_value(0)
+            self._progress_scale_zero()
             self.progressScale.set_sensitive(False)
             if self.playlist is not None:
                 currentTrack = self.playlist.get_path(self.playlist.get_iter_first())
@@ -386,7 +389,7 @@ class Player(GObject.GObject):
             # Stop playback
             self.stop()
             self.playBtn.set_image(self._playImage)
-            self.progressScale.set_value(0)
+            self._progress_scale_zero()
             self.progressScale.set_sensitive(False)
             self.emit('playback-status-changed')
 
@@ -587,36 +590,23 @@ class Player(GObject.GObject):
 
     @log
     def load(self, media):
-        self.progressScale.set_value(0)
+        self._progress_scale_zero()
         self._set_duration(media.get_duration())
-        self.songTotalTimeLabel.set_label(self.seconds_to_string(media.get_duration()))
+        self.songTotalTimeLabel.set_label(
+            utils.seconds_to_string(media.get_duration()))
         self.progressScale.set_sensitive(True)
 
         self.playBtn.set_sensitive(True)
         self._sync_prev_next()
 
-        artist = _("Unknown Artist")
-        try:
-            assert media.get_artist() is not None
-            artist = media.get_artist()
-        except:
-            pass
-        finally:
-            self.artistLabel.set_label(artist)
-            self._currentArtist = artist
+        artist = utils.get_artist_name(media)
+        self.artistLabel.set_label(artist)
+        self._currentArtist = artist
 
-        album = _("Unknown Album")
-        try:
-            assert media.get_album() is not None
-            album = media.get_album()
-        except:
-            self._currentAlbum = album
+        self.coverImg.set_from_surface(self._loading_icon_surface)
+        self.cache.lookup(media, ArtSize.xsmall, self._on_cache_lookup, None)
 
-        self.coverImg.set_from_pixbuf(self._noArtworkIcon)
-        self.cache.lookup(
-            media, ART_SIZE, ART_SIZE, self._on_cache_lookup, None, artist, album)
-
-        self._currentTitle = AlbumArtCache.get_media_title(media)
+        self._currentTitle = utils.get_media_title(media)
         self.titleLabel.set_label(self._currentTitle)
 
         self._currentTimestamp = int(time.time())
@@ -656,7 +646,7 @@ class Player(GObject.GObject):
         nextSong = self.playlist.get_value(_iter, self.playlistField)
         url = self.playlist.get_value(_iter, 5).get_url()
 
-        # Skip remote tracks discovery
+        # Skip remote songs discovery
         if url.startswith('http://') or url.startswith('https://'):
             return False
         elif status == DiscoveryStatus.PENDING:
@@ -667,10 +657,9 @@ class Player(GObject.GObject):
         return False
 
     @log
-    def _on_cache_lookup(self, pixbuf, path, data=None):
-        if pixbuf is not None:
-            self.coverImg.set_from_pixbuf(pixbuf)
-        self.emit('thumbnail-updated', path)
+    def _on_cache_lookup(self, surface, data=None):
+        self.coverImg.set_from_surface(surface)
+        self.emit('thumbnail-updated')
 
     @log
     def play(self):
@@ -694,17 +683,15 @@ class Player(GObject.GObject):
             t = Thread(target=self.update_now_playing_in_lastfm, args=(media.get_url(),))
             t.setDaemon(True)
             t.start()
-        if not self.timeout:
-            self.timeout = GLib.timeout_add(1000, self._update_position_callback)
+        if not self.timeout and self.progressScale.get_realized():
+            self._update_timeout()
 
         self.emit('playback-status-changed')
         self.emit('playing-changed')
 
     @log
     def pause(self):
-        if self.timeout:
-            GLib.source_remove(self.timeout)
-            self.timeout = None
+        self._remove_timeout()
 
         self.player.set_state(Gst.State.PAUSED)
         self.emit('playback-status-changed')
@@ -712,9 +699,7 @@ class Player(GObject.GObject):
 
     @log
     def stop(self):
-        if self.timeout:
-            GLib.source_remove(self.timeout)
-            self.timeout = None
+        self._remove_timeout()
 
         self.player.set_state(Gst.State.NULL)
         self.emit('playing-changed')
@@ -728,7 +713,7 @@ class Player(GObject.GObject):
             return True
 
         self.stop()
-        self.currentTrack = self._get_next_track()
+        self.currentTrack = self.nextTrack
 
         if self.currentTrack and self.currentTrack.valid():
             self.currentTrackUri = self.playlist.get_value(
@@ -745,7 +730,7 @@ class Player(GObject.GObject):
 
         position = self.get_position() / 1000000
         if position >= 5:
-            self.progressScale.set_value(0)
+            self._progress_scale_zero()
             self.on_progress_scale_change_value(self.progressScale)
             return
 
@@ -764,8 +749,11 @@ class Player(GObject.GObject):
         else:
             self.set_playing(True)
 
+    # FIXME: set the discovery field to 11 to be safe, but for some
+    # models it is 12.
     @log
-    def set_playlist(self, type, id, model, iter, field, discovery_status_field):
+    def set_playlist(self, type, id, model, iter, field,
+                     discovery_status_field=11):
         self.stop()
 
         old_playlist = self.playlist
@@ -814,6 +802,9 @@ class Player(GObject.GObject):
         self.titleLabel = self._ui.get_object('title')
         self.artistLabel = self._ui.get_object('artist')
         self.coverImg = self._ui.get_object('cover')
+        self.coverImg.set_property("width-request", ArtSize.xsmall.width)
+        self.coverImg.set_property("height-request", ArtSize.xsmall.height)
+
         self.duration = self._ui.get_object('duration')
         self.repeatBtnImage = self._ui.get_object('playlistRepeat')
 
@@ -832,35 +823,106 @@ class Player(GObject.GObject):
         self.progressScale.connect('button-press-event', self._on_progress_scale_event)
         self.progressScale.connect('value-changed', self._on_progress_value_changed)
         self.progressScale.connect('button-release-event', self._on_progress_scale_button_released)
+        self.progressScale.connect('change-value', self._on_progress_scale_seek)
+        self._ps_draw = self.progressScale.connect('draw',
+            self._on_progress_scale_draw)
+        self._seek_timeout = None
+        self._old_progress_scale_value = 0.0
+
+    def _on_progress_scale_seek_finish(self, value):
+        """Prevent stutters when seeking with infinitesimal amounts"""
+        self._seek_timeout = None
+        round_digits = self.progressScale.get_property('round-digits')
+        if self._old_progress_scale_value != round(value, round_digits):
+            self.on_progress_scale_change_value(self.progressScale)
+            self._old_progress_scale_value = round(value, round_digits)
+        return False
+
+    def _on_progress_scale_seek(self, scale, scroll_type, value):
+        """Smooths out the seeking process
+
+        Called every time progress scale is moved. Only after a seek has been
+        stable for 100ms, we play the song from its location.
+        """
+        if self._seek_timeout:
+            GLib.source_remove(self._seek_timeout)
+
+        Gtk.Range.do_change_value(scale, scroll_type, value)
+        self._seek_timeout = GLib.timeout_add(100,
+                                              self._on_progress_scale_seek_finish,
+                                              value)
+        return True
 
     @log
     def _on_progress_scale_button_released(self, scale, data):
-        self.on_progress_scale_change_value(self.progressScale)
+        if self._seek_timeout:
+            GLib.source_remove(self._seek_timeout)
+            self._on_progress_scale_seek_finish(self.progressScale.get_value())
+
         self._update_position_callback()
-        self.player.set_state(self._lastState)
-        self.timeout = GLib.timeout_add(1000, self._update_position_callback)
         return False
 
     def _on_progress_value_changed(self, widget):
         seconds = int(self.progressScale.get_value() / 60)
-        self.songPlaybackTimeLabel.set_label(self.seconds_to_string(seconds))
+        self.songPlaybackTimeLabel.set_label(utils.seconds_to_string(seconds))
         return False
 
     @log
     def _on_progress_scale_event(self, scale, data):
-        self._lastState = self.player.get_state(1)[1]
-        self.player.set_state(Gst.State.PAUSED)
+        self._remove_timeout()
+        self._old_progress_scale_value = self.progressScale.get_value()
+        return False
+
+    def _on_progress_scale_draw(self, cr, data):
+        self._update_timeout()
+        self.progressScale.disconnect(self._ps_draw)
+        return False
+
+    def _update_timeout(self):
+        """Update the duration for self.timeout and self._seconds_timeout
+
+        Sets the period of self.timeout to a value small enough to make the
+        slider of self.progressScale move smoothly based on the current song
+        duration and progressScale length.  self._seconds_timeout is always set
+        to a fixed value, short enough to hide irregularities in GLib event
+        timing from the user, for updating the songPlaybackTimeLabel.
+        """
+        # Don't run until progressScale has been realized
+        if self.progressScale.get_realized() == False:
+            return
+
+        # Update self.timeout
+        width = self.progressScale.get_allocated_width()
+        padding = self.progressScale.get_style_context().get_padding(
+            Gtk.StateFlags.NORMAL)
+        width -= padding.left + padding.right
+        success, duration = self.player.query_duration(Gst.Format.TIME)
+        timeout_period = 1000
+        if success:
+            timeout_period = min(1000 * (duration / 10**9) // width, 1000)
+
+        if self.timeout:
+            GLib.source_remove(self.timeout)
+        self.timeout = GLib.timeout_add(
+            timeout_period, self._update_position_callback)
+
+        # Update self._seconds_timeout
+        if not self._seconds_timeout:
+            self.seconds_period = 1000
+            self._seconds_timeout = GLib.timeout_add(
+                self.seconds_period, self._update_seconds_callback)
+
+    def _remove_timeout(self):
         if self.timeout:
             GLib.source_remove(self.timeout)
             self.timeout = None
-        return False
+        if self._seconds_timeout:
+            GLib.source_remove(self._seconds_timeout)
+            self._seconds_timeout = None
 
-    def seconds_to_string(self, duration):
-        seconds = duration
-        minutes = seconds // 60
-        seconds %= 60
-
-        return '%i:%02i' % (minutes, seconds)
+    def _progress_scale_zero(self):
+        self.progressScale.set_value(0)
+        self._on_progress_value_changed(None)
 
     @log
     def _on_play_btn_clicked(self, btn):
@@ -887,8 +949,6 @@ class Player(GObject.GObject):
     @log
     def scrobble_song(self, url):
         # Update playlists
-        playlists.update_playcount(url)
-        playlists.update_last_played(url)
         playlists.update_all_static_playlists()
 
         if self.last_fm:
@@ -948,13 +1008,23 @@ class Player(GObject.GObject):
         position = self.player.query_position(Gst.Format.TIME)[1] / 1000000000
         if position > 0:
             self.progressScale.set_value(position * 60)
-            self.played_seconds += 1
+        self._update_timeout()
+        return False
+
+    def _update_seconds_callback(self):
+        self._on_progress_value_changed(None)
+
+        position = self.player.query_position(Gst.Format.TIME)[1] / 10**9
+        if position > 0:
+            self.played_seconds += self.seconds_period / 1000
             try:
                 percentage = self.played_seconds / self.duration
                 if not self.scrobbled and percentage > 0.4:
                     current_media = self.get_current_media()
                     self.scrobbled = True
                     if current_media:
+                        grilo.bump_play_count(self.get_current_media())
+                        grilo.set_last_played(current_media)
                         just_played_url = self.get_current_media().get_url()
                         t = Thread(target=self.scrobble_song, args=(just_played_url,))
                         t.setDaemon(True)
@@ -1004,7 +1074,7 @@ class Player(GObject.GObject):
 
     @log
     def Stop(self):
-        self.progressScale.set_value(0)
+        self._progress_scale_zero()
         self.progressScale.set_sensitive(False)
         self.playBtn.set_image(self._playImage)
         self.stop()
